# THEMER #
# elevate.py #
# COPYRIGHT (c) barneygale 2018- #
# Taken from https://github.com/barneygale/elevate, with notice to COPYING.txt #

# Imports #
import errno
import os
import sys
try:
    from shlex import quote
except ImportError:
    from pipes import quote


def quote_shell(args):
    return " ".join(quote(arg) for arg in args)


def quote_applescript(string):
    charmap = {
        "\n": "\\n",
        "\r": "\\r",
        "\t": "\\t",
        "\"": "\\\"",
        "\\": "\\\\",
    }
    return '"%s"' % "".join(charmap.get(char, char) for char in string)


def elevate(show_console=True, graphical=True):
    if os.getuid() == 0:
        return

    args = [sys.executable] + sys.argv
    commands = []

    if graphical:
        if sys.platform.startswith("darwin"):
            commands.append([
                "osascript",
                "-e",
                "do shell script %s "
                "with administrator privileges "
                "without altering line endings"
                % quote_applescript(quote_shell(args))])

    commands.append(["sudo"] + args)

    for args in commands:
        try:
            os.execlp(args[0], *args)
        except OSError as e:
            if e.errno != errno.ENOENT or args[0] == "sudo":
                raise



# THEMER #
# apps.py #
# COPYRIGHT (c) Jaidan 2022- #

# Imports #
from os import getuid
from glob import glob
from plistlib import loads
from themer.elevate import elevate
from themer.logger import logger

async def plist(app_name: str, apps: list[dict]):
    '''Read an app's plist.'''
    app_obj = None
    for app in apps:
        if app.get('name') == app_name:
            app_obj = app
    if app_obj is None:
        return None
    try:
        with open(app_obj.get('path') + '/Contents/Info.plist', 'rb') as f:
            try:
                plist = loads(f.read())
            except:
                logger.error('Could not parse {}\'s Info.plist.'.format(app_name))
                exit(1)
    except:
        logger.error('Error reading {}\'s Info.plist'.format(app_name))
        exit(1)
    try:
        return_value = {
            'CFBundleName': plist.get('CFBundleName'),
            'CFBundleIconFile': plist.get('CFBundleIconFile'),
            'CFBundleTypeIconFile': plist.get('CFBundleTypeIconFile'),
            'CFBundleExecutable': plist.get('CFBundleExecutable'),
            'AppPath': app_obj.get('path')
        }
    except:
        logger.error('Error reading {}\'s Info.plist'.format(app_name))
        exit(1)
    
    return return_value

def get_apps() -> list[dict]:
    '''Get all installed apps.'''
    try:
        apps = []
        if getuid() != 0:
            logger.info('Requesting root privileges to read and modify system applications.')
            elevate()
        for app in glob('/System/Applications/*.app'): apps.append({'name': app.split('/')[-1].replace('.app', ''), 'path': app})
        for app in glob('/Applications/*.app'): apps.append({'name': app.split('/')[-1].replace('.app', ''), 'path': app})
        for app in glob('/Applications/*/*.app'): apps.append({'name': app.split('/')[-1].replace('.app', ''), 'path': app})
        return apps
    except:
        logger.error('Could not get app list.')
        exit(1)


# THEMER #
# logger.py #
# COPYRIGHT (c) Jaidan 2022- #

# Imports #
import logging
import radium as r

class Logger:
    def __init__(self):
        self.logger = logging.Logger(__name__)
        self.logger.setLevel(logging.INFO)
        self.logger.addHandler(r.Radium)

logger = Logger().logger


# THEMER #
# main.py #
# COPYRIGHT (c) Jaidan 2022- #

# Imports #
import zipfile
from os import (system, unlink)
from themer.logger import logger
from themer.recovery import (backup, enact_backup)
from themer.typedef import Theme
from yaml import (FullLoader, load)

def unpack(path: str) -> dict:
    '''Unpacks a theme file and provides all of it's data.'''
    try:
        # Open zip file #
        with zipfile.ZipFile(path, 'r') as zip:
            # Initialize theme object #
            theme_obj = None
            # Loop through files #
            for file in zip.filelist:
                # If the file ends with .yaml #
                if file.filename.endswith('.yml'):
                    # Get the directory in the zip #
                    directory = file.filename.split('/')[0]
                    # Read the yaml #
                    yaml_obj = load(zip.read(file).decode(
                        'utf-8'), Loader=FullLoader)
                    # Assign the theme object #
                    theme_obj = Theme(yaml_obj)
                    # Loop through icons #
                    for icon in theme_obj.icons:
                        try:
                            # Read icon data #
                            icon.data = zip.read(directory + '/' + icon.img)
                        except:
                            logger.error(
                                'Could not read {}\'s icon.'.format(icon.name))
                            exit(1)
            if theme_obj is None:
                logger.logger('Could not find theme manifest.')
                exit(1)
            else:
                # Return #
                return {'theme': theme_obj, 'zip': zip}
    except zipfile.BadZipFile:
        logger.error('Invalid zip file.')
        exit(1)
            
def refresh_cache():
    system('killall Finder')
    system('killall Dock')
            
def activate_theme(path: str):
    '''Activate a theme.'''
    # Get theme package #
    theme_pack = unpack(path)
    # Parse to object #
    theme: Theme = theme_pack.get('theme')
    # Print theme name #
    logger.info('Activating \'{}\'.'.format(theme.name))
    # Loop through icons #
    for icon in theme.icons:
        if icon.plist.icon_asset is None:
            continue
        # Make icon backup #
        backup(icon.plist)
        # Replace icon #
        with open(str(icon.plist.resource_path + '/' + icon.plist.icon_asset + '.icns'), 'wb') as img:
            img.write(icon.data)
            system('touch {}'.format(str(icon.plist.path.replace(' ', '\ '))))
            img.close()
    # Clear cache #
    logger.info('Refreshing cache...')
    refresh_cache()
    logger.info('Successfully activated \'{}\'.'.format(theme.name))
    
def deactivate_theme(path: str):
    '''Deactivate a theme.'''
    # Get theme package #
    theme_pack = unpack(path)
    # Parse to object #
    theme: Theme = theme_pack.get('theme')
    # Print theme name #
    logger.info('Deactivating \'{}\'.'.format(theme.name))
    # Loop through icons #
    for icon in theme.icons:
        if icon.plist.icon_asset is None:
            continue
        # Remove icon file #
        unlink(str(icon.plist.resource_path + '/' + icon.plist.icon_asset + '.icns'))
        # Enact icon backup #
        enact_backup(icon.plist)
        # Clear cache #
        system('touch {}'.format(str(icon.plist.path.replace(' ', '\ '))))
    # Clear cache #
    logger.info('Refreshing cache...')
    refresh_cache()
    logger.info('Successfully deactivated \'{}\'.'.format(theme.name))


# THEMER #
# types.py #
# COPYRIGHT (c) Jaidan 2022- #

# Imports #
from asyncio import run
from themer.apps import (get_apps, plist)

class Plist():
    def __init__(self, app_name: str, app_list: list[dict]):
        # Plist #
        plist_obj = run(plist(app_name, app_list))
        # App name #
        self.name = app_name
        # App icon asset #
        self.icon_asset = str(plist_obj.get(
            'CFBundleIconFile')).replace('.icns', '') if plist_obj else None
        # App path #
        self.path = str(plist_obj.get('AppPath')) if plist_obj else None
        # App resource path #
        self.resource_path = self.path + '/Contents/Resources' if self.path else None
        # App executable
        self.executable = plist_obj.get('CFBundleExecutable') if plist_obj else None

class Icon():
    def __init__(self, dict: dict, apps: list[dict]):
        # App name #
        self.name = dict.get('name')
        # App icon relative path #
        self.img = dict.get('img')
        # App icon data #
        self.data = None
        # Plist #
        self.plist = Plist(dict.get('name'), apps)

class Theme():
    def __init__(self, dict: dict):
        # Theme name #
        self.name = dict.get('name')
        # Icons object #
        self.icons: list[Icon] = []
        # Icons #
        app_list = get_apps()
        for icon in dict.get('icons'):
            self.icons.append(Icon(icon, app_list))



# THEMER #
# apps.py #
# COPYRIGHT (c) Jaidan 2022- #

# Imports #
from genericpath import exists
from os import remove
from themer.logger import logger
from themer.typedef import Plist

def __icon_backup_name(plist: Plist) -> str: return str(plist.resource_path + '/' + plist.name + '-' + plist.icon_asset + '-ORIGINAL.icns')

def check(plist: Plist) -> bool: 
    '''Check if an app's original icon has been backed up.'''
    try:
        return exists(__icon_backup_name(plist))
    except:
        logger.error('Could not check if {}\'s original icon has been backed up. ({})'.format(plist.name, plist.path))
        exit(1)

def backup(plist: Plist):
    '''Back up an app's original icon.'''
    try:
        # Check if it's already backed up #
        if check(plist): return
        # Get the app's icon #
        data = open(str(plist.resource_path + '/' + plist.icon_asset + '.icns'), 'rb').read()
        # Write to backup #
        open(__icon_backup_name(plist), 'wb').write(data)
        # Return #
        return
    except:
        logger.error('Unable to back up original icon for {}. ({})'.format(plist.name, plist.path))
        exit(1)

def enact_backup(plist: Plist):
    '''Recover an app's original icon.'''
    try:
        # Check #
        if not check(plist): return
        with open(str(plist.resource_path + '/' + plist.icon_asset + '.icns'), 'wb') as f:
            # Get backup #
            data = open(__icon_backup_name(plist), 'rb').read()
            # Write #
            f.write(data)
            # Delete backup #
            remove(__icon_backup_name(plist))
            # Close #
            f.close()
        # Return #
        return
    except:
        logger.error('Could not reset to original icon for {}. ({})'.format(plist.name, plist.path))
        exit(1)



from setuptools import setup, find_packages
from io import open
from os import path

import pathlib
# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# automatically captured required modules for install_requires in requirements.txt
with open(path.join(HERE, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if ('git+' not in x) and (
    not x.startswith('#')) and (not x.startswith('-'))]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs
                    if 'git+' not in x]

setup(
    name='pythemer',
    description='A WIP theme engine for macOS, written in Python.',
    version='1.0.0',
    packages=find_packages(),  # list of all packages
    install_requires=install_requires,
    python_requires='>=2.7',  # any python greater than 2.7
    entry_points='''
        [console_scripts]
        themer=themer.__main__:main
    ''',
    author="Jaidan",
    keywords="theme, themes, macos, mac, icon",
    long_description=README,
    long_description_content_type="text/markdown",
    license='JL',
    url='https://github.com/pythemer/themer',
    download_url='https://github.com/pythemer/themer/archive/1.0.0.tar.gz',
    dependency_links=dependency_links,
    author_email='ajaidan0@gmail.com',
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ]
)



# THEMER #
# __main__.py #
# COPYRIGHT (c) Jaidan 2022- #

# Imports #
import click
from themer.theme import (activate_theme, deactivate_theme)

__author__ = "Jaidan"

@click.group()
def main():
    '''A WIP theme engine for macOS, written in Python.'''
    pass

@main.command()
@click.argument('path', type=click.Path(exists=True))
def activate(path):
    '''Activate a theme.'''
    activate_theme(path)

@main.command()
@click.argument('path', type=click.Path(exists=True))
def deactivate(path):
    '''Deactivate a theme.'''
    deactivate_theme(path)

if __name__ == "__main__":
    main()